#!/usr/bin/env python
"""

Copyright (c) 2014 - 2015 Men & Mice

Permission is hereby granted, free of charge, to any person obtaining a copy of
this software and associated documentation files (the "Software"), to deal in
the Software without restriction, including without limitation the rights to
use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies
of the Software, and to permit persons to whom the Software is furnished to do
so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


This module requires the requests library (http://docs.python-requests.org/en/latest/). 
If you do not have it installed you will need to install it. To install simply run:
    $ pip install requests
or, with easy_install:
    $ easy_install requests

"""
from __future__ import print_function	# Prepare for python 3.1
import requests
from requests import Request, Session
import json
import random

from logging import getLogger
LOG = getLogger(__name__)

class JSONClient(object):
    """
    A simple class that knowns how to communicate with Men & Mice Web Service using 
    the JSON-RPC protocol.

    It overloads the __getattr__ method so that all unknown functions are assumed to be
    RPC requests and the arguments are assumed to be a RPC data.

    Args:
        proxy:          Name or an address of the a Men & Mice Web Service.
                        If none is provided the proxy is assumed to be located
                        on the same host as the Men & Mice Central Service.
        sessionid:      A valid session ID. Can be used if you have already have a 
                        valid session ID if so then you don't have to call Login().

    Returns:
        A valid JSONClient object that can be used to log in to Men & Mice Central
        Service and perform various functions. For a list of possible functions see:
        https://docs.menandmice.com/current/SOAP+API

    """
    def __init__(self, sessionid='', proxy=None):
        self._sess = Session()
        self.sessionid = sessionid
        self._proxy = proxy
        if self._proxy:
            self._url = "http://" + self._proxy + "/_mmwebext/mmwebext.dll?json"

    def __getstate__(self):
        return {'_proxy': self._proxy, 'sessionID': self.sessionid}

    def __setstate__(self, state):
        self.__init__(sessionid=state['sessionID'], proxy=state['_proxy'])

    def __getattr__(self, name):
        if name.startswith("_"):
            raise AttributeError
        def method(*args, **kwargs):
            request_args = kwargs
            if args:
                for arg in args:
                    if type(arg) is dict:
                        request_args.update(arg)
            return self.send_request(name, request_args)
        return method

    def send_request(self, request_name, request_data=None):
        """
        Sends a JSON-RPC request to the connected Men & Mice Central Service.

        Args:
            request_name:   The name of the request.
            request_data:   The request data.

        Returns:
            The result data as a json object

        """
        if not self._proxy:
            raise Exception('error: no proxy name has been provided')
        if request_data == None:
            request_data = {}
        my_msgid = random.randint(1, 9999)
        if not 'session' in request_data and self.sessionid:
            request_data['session'] = self.sessionid
        data = '{"jsonrpc": "2.0", "method": "' + request_name + '", "params": ' + json.dumps(request_data) + ', "id": '+ str(my_msgid) + '}'
        #LOG.info(">>>>> <" + data + ">")
        resp = self._sess.post(self._url, headers={'content-type':'application/json', 'accept':'application/json'}, data=data)
        if resp.status_code != 200:    # fatal
            raise Exception("error: unexpected result [" + str(resp.status_code) + '] ' + resp.reason)
        # it differs between versions of request what json actually is. 2.7 it is a dict in 3.X it is a method
        if isinstance(resp.json, dict):
            data = resp.json
        else:
            data = resp.json()
        #LOG.info("<<<<< <" + json.dumps(data) + ">")
        if 'error' in data:
            raise Exception('error: ' + str(data['error'].get('code', '0')) + ' ' + data['error'].get('message', ''))
        if not 'id' in data or data['id'] != my_msgid:    # fatal
            raise Exception('error: incorrect message id: ' + str(data.get('id', '0')))
        if not 'result' in data:    # fatal
            raise Exception('error: Missing result')
        return data['result']

    def send_plain_request(self, request_data):
        """
        Sends a request to the connected Men & Mice Central Service.

        Args:
            request_data:   The request data. Note in order to have any meaning to
                            the web service, it has to be a legal JSON-RPC 2.0 string
                            (see: http://www.jsonrpc.org/specification)
        Returns:
            The result data as a plain text

        """
        if not self._proxy:
            raise Exception('error: no proxy name has been provided')
        #LOG.info(">>>>> <" + request_data + ">")
        resp = self._sess.post(self._url, headers={'content-type':'application/json', 'accept':'application/json'}, data=request_data)
        if resp.status_code != 200:    # fatal
            raise Exception("error: unexpected result [" + str(resp.status_code) + '] ' + resp.reason)
        data = resp.text
        #LOG.info("<<<<< <" + data + ">")
        return data

    def Login(self, proxy=None, server='localhost', username='administrator', password='administrator'):
        # pylint: disable=invalid-name
        """
        Log in to a Men & Mice Central service. If succesful it will get a valid session
        ID that will be kept and used in all functions.

        Args:
            proxy:          Name or an address of the a Men & Mice Web Service.
                            If none is provided the proxy is assumed to be located
                            on the same host as the Men & Mice Central Service.
            server:         Name or an address of the a Men & Mice Central Service.
            username:       The name of the user to be logged in as.
            password:       The password for the user to be logged in.

        Returns:
            
        """
        if proxy:
            self._proxy = proxy
        # if there is no proxy defined, use server
        if not self._proxy:
            self._proxy = server
        self._url = "http://" + self._proxy + "/_mmwebext/mmwebext.dll?json"
        self.sessionid = ''
        # do the actual login
        resp = self.send_request('Login', {'server': server, 'loginName': username, 'password': password})
        if 'session' in resp:
            self.sessionid = resp['session']

    def Logout(self):
        # pylint: disable=invalid-name
        """
        Logs out of the current session and invalidates the current session ID.

        """
        self.send_request('Logout')
        self.sessionid = ''



#
# Example code
#
def main():
    client = JSONClient()
    client.Login(server='192.168.91.130')
    # client.Login(server='ca.dev.lab')

    servers = client.GetDNSServers()
    print(servers)
    zones = client.GetDNSZones()
    print(zones)
    records = client.GetDNSRecords(dnsZoneRef='{#4-#70}')
    print(records)

if __name__ == '__main__':
    main()
